# GlycoGenius: Glycomics Data Analysis Tool
# Copyright (C) 2023 by Hector Franco Loponte
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or 
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. It is accessible within the program files
# or by typing 'license' after running it stand-alone in the terminal
# by typing 'glycogenius'. If not, see <https://www.gnu.org/licenses/>.

import pathlib
import importlib
import_path = str(pathlib.Path(__file__).parent.resolve())
for i_i, i in enumerate(import_path):
    if i == "\\":
        import_path = import_path[:i_i]+"/"+import_path[i_i+1:]
        
#Absolute import of General_Functions
spec1 = importlib.util.spec_from_file_location("General_Functions", import_path+'/General_Functions.py')
General_Functions = importlib.util.module_from_spec(spec1)
spec1.loader.exec_module(General_Functions)

from pyteomics import mzxml, mzml, mass, auxiliary
from itertools import combinations_with_replacement
from scipy.signal import savgol_filter
from scipy.sparse.linalg import splu
from scipy import sparse
from statistics import mean
from re import split
from math import inf, atan, pi
import numpy
import sys
import datetime

##---------------------------------------------------------------------------------------
##File accessing-associated functions (these are functions that deal with the data in
##some way. They make vast use of general functions and require a library to work).

class make_mzxml(object):
    '''A wrapper that takes the output of pyteomics mzML parser and converts it to
    the mzXML pyteomics parser standard to be used within the script. Allows for full
    support of mzML.
    
    Parameters
    ----------
    object : string
        A string containing the path to the mzML file.
        
    Uses
    ----
    pyteomics.mzml.MzML : generator
        Indexes the mzML file into a generator, allowing you to parse the file for
        analysis.
        
    Returns
    -------
    standard mzML output
        If class is iterated, returns regular mzml.MzML output.
    
    dict
        If class is queried for a single index, returns the dictionary of the
        converted output (mzML -> mzXML)
        
    data : list
        If class is queried for slices, returns a list of dictionaries containing
        the converted output (mzML -> mzXML)
    '''
    def __init__(self,it):
        self.it = mzml.MzML(it)
    def __iter__(self):
        return self.it
    def __getitem__(self,index):
        if type(index) == int:
            pre_data = self.it[index]
            if pre_data['ms level'] == 2:
                if float(self.it[-1]['scanList']['scan'][0]['scan start time']) > 210: #210 scan time should allow for the correct evaluation of scan time being in seconds or minutes for every run that lasts between 3.5 minutes and 3.5 hours
                    return {'num': pre_data['id'].split('=')[-1], 'retentionTime': float(pre_data['scanList']['scan'][0]['scan start time'])/60, 'msLevel': pre_data['ms level'], 'm/z array': pre_data['m/z array'], 'intensity array': pre_data['intensity array'], 'precursorMz': [{'precursorMz': pre_data['precursorList']['precursor'][0]['isolationWindow']['isolation window target m/z']}]}
                else:
                    return {'num': pre_data['id'].split('=')[-1], 'retentionTime': float(pre_data['scanList']['scan'][0]['scan start time']), 'msLevel': pre_data['ms level'], 'm/z array': pre_data['m/z array'], 'intensity array': pre_data['intensity array'], 'precursorMz': [{'precursorMz': pre_data['precursorList']['precursor'][0]['isolationWindow']['isolation window target m/z']}]}
            else:
                if float(self.it[-1]['scanList']['scan'][0]['scan start time']) > 210:
                    return {'num': pre_data['id'].split('=')[-1], 'retentionTime': float(pre_data['scanList']['scan'][0]['scan start time'])/60, 'msLevel': pre_data['ms level'], 'm/z array': pre_data['m/z array'], 'intensity array': pre_data['intensity array']}
                else:
                    return {'num': pre_data['id'].split('=')[-1], 'retentionTime': float(pre_data['scanList']['scan'][0]['scan start time']), 'msLevel': pre_data['ms level'], 'm/z array': pre_data['m/z array'], 'intensity array': pre_data['intensity array']}
        else:
            first_index = str(index).split('(')[1].split(', ')[0]
            if first_index != 'None':
                first_index = int(first_index)
            else:
                first_index = 0
            last_index = str(index).split('(')[1].split(', ')[1]
            if last_index != 'None':
                last_index = int(last_index)
            else:
                last_index = len(self.it)
            data = []
            for index in range(first_index, last_index):
                if self.it[index]['ms level'] == 2:
                    if float(self.it[-1]['scanList']['scan'][0]['scan start time']) > 210:
                        data.append({'num': self.it[index]['id'].split('=')[-1], 'retentionTime': float(self.it[index]['scanList']['scan'][0]['scan start time'])/60, 'msLevel': self.it[index]['ms level'], 'm/z array': self.it[index]['m/z array'], 'intensity array': self.it[index]['intensity array'], 'precursorMz': [{'precursorMz': self.it[index]['precursorList']['precursor'][0]['isolationWindow']['isolation window target m/z']}]})
                    else:
                        data.append({'num': self.it[index]['id'].split('=')[-1], 'retentionTime': float(self.it[index]['scanList']['scan'][0]['scan start time']), 'msLevel': self.it[index]['ms level'], 'm/z array': self.it[index]['m/z array'], 'intensity array': self.it[index]['intensity array'], 'precursorMz': [{'precursorMz': self.it[index]['precursorList']['precursor'][0]['isolationWindow']['isolation window target m/z']}]})
                else:
                    if float(self.it[-1]['scanList']['scan'][0]['scan start time']) > 210:
                        data.append({'num': self.it[index]['id'].split('=')[-1], 'retentionTime': float(self.it[index]['scanList']['scan'][0]['scan start time'])/60, 'msLevel': self.it[index]['ms level'], 'm/z array': self.it[index]['m/z array'], 'intensity array': self.it[index]['intensity array']})
                    else:
                        data.append({'num': self.it[index]['id'].split('=')[-1], 'retentionTime': float(self.it[index]['scanList']['scan'][0]['scan start time']), 'msLevel': self.it[index]['ms level'], 'm/z array': self.it[index]['m/z array'], 'intensity array': self.it[index]['intensity array']})
            return data
        
def eic_from_glycan(files,
                    glycan,
                    glycan_info,
                    ms1_indexes,
                    rt_interval,
                    tolerance,
                    min_isotops,
                    noise,
                    avg_noise,
                    max_charges,
                    fast_iso,
                    verbose = False):
    '''Generates a very processed EIC for each adduct of each glycan of each sample.
    Removes non-monoisotopic peaks, check charges and calculates multiple quality scoring
    data.

    Parameters
    ----------
    files : list
        A list of generators obtained from the function pyteomics.mzxml.MzXML().

    glycan_info : dict
        A dictionary containing multiple glycan info, including the needed 'Adducts_mz'
        key required for this function. Generated by full_glycans_library().

    ms1_indexes : dict
        A dictionary where each key is the ID of a file and each value is a list
        containing the indexes of all MS1 scans in the file.
        
    rt_interval : tuple
        A tuple where the first index contains the beggining time of the retention time
        interval you wish to analyze and the second contains the end time.
        
    tolerance : tuple
        First index contains the unit of the tolerance and the second one is the value of 
        that unit.
        
    min_isotops : int
        The minimum amount of isotopologues required to consider an RT mz peak valid.
        
    noise : list
        A list containing the calculated noise level of each sample.
    
    avg_noise : float
        Fallback average noise in case local noise can't be calculated.
        
    max_charges : int
        The maximum amount of charges the queried mz should have.
        
    verbose : boolean
        Allows to print to file debugging information of each EIC traced so that you
        may find out why some specific pattern or behavior is ocurring.

    Uses
    ----
    General_Functions.form_to_charge : int
        Converts adducts formula into raw charge.
        
    pyteomics.mass.calculate_mass(*args, **kwargs) : float
        Calculates the monoisotopic mass of a polypeptide defined by a sequence string,
        parsed sequence, chemical formula or Composition object.
        
    General_Functions.form_to_comp(string) : dict
        Separates a molecular formula or monosaccharides composition of glycans into a
        dictionary with each atom/monosaccharide as a key and its amount as value.
        
    General_Functions.h_mass : float
        The mass of Hydrogen-1.
        
    General_Functions.calculate_ppm_diff : float
        Calculates the PPM difference between a mz and a target mz.
        
    numpy.corrcoef : matrix
        Return Pearson product-moment correlation coefficients.

    Returns
    -------
    data : dict
        A dictionary with keys for each adduct combo of a given glycan, which value is a
        dict with keys for each file id, at which value is a list of lists, with the
        first one being the rt array and the second one the processed int array.
        
    ppm_info : dict
        A dictionary with keys for each adduct combo of a given glycan, which value is a
        dict with keys for each file id, at which each value corresponds to a given 
        calculated ppm difference for the retention time at the same index.
        
    iso_fitting_quality : dict
        A dictionary with keys for each adduct combo of a given glycan, which value is a
        dict with keys for each file id, at which each value corresponds to a given 
        calculated isotopic fitting score for the retention time at the same index.
        
    verbose_info : list
        A list containing the conclusion of the processing for each retention time of each
        chromatogram processed, for each file.
        
    raw_data : dict
        A dictionary with keys for each adduct combo of a given glycan, which value is a
        dict with keys for each file id, at which value is a list of lists, with the
        first one being the rt array and the second one the raw int array.
    '''
    data = {}
    ppm_info = {}
    iso_fitting_quality = {}
    verbose_info = []
    raw_data = {}
    for i in glycan_info['Adducts_mz']:
        if verbose:
            print('Adduct: '+str(i)+" mz: "+str(glycan_info['Adducts_mz'][i]))
            verbose_info.append('Adduct: '+str(i)+" mz: "+str(glycan_info['Adducts_mz'][i]))
        adduct_mass = mass.calculate_mass(composition=General_Functions.form_to_comp(i))
        adduct_charge = General_Functions.form_to_charge(i)
        ppm_info[i] = {}
        iso_fitting_quality[i] = {}
        data[i] = {}
        raw_data[i] = {}
        for j_j, j in enumerate(files):
            if verbose:
                print("--Drawing EIC for Sample: "+str(j_j))
                verbose_info.append("--Sample: "+str(j_j))
            ppm_info[i][j_j] = []
            iso_fitting_quality[i][j_j] = []
            data[i][j_j] = [[], []]
            raw_data[i][j_j] = [[], []]
            for k_k, k in enumerate(ms1_indexes[j_j]):
                iso_fitting_quality[i][j_j].append(0.0)
                ppm_info[i][j_j].append(inf)
                not_good = False
                found = False
                rt = j[k]['retentionTime']
                data[i][j_j][0].append(rt)
                raw_data[i][j_j][0].append(rt)
                if (j[k]['retentionTime'] < rt_interval[0] or j[k]['retentionTime'] > rt_interval[1]):
                    data[i][j_j][1].append(0.0)
                    raw_data[i][j_j][1].append(0.0)
                    continue
                else:   
                    if verbose:
                        verbose_info.append("----RT: "+str(rt))
                    l_l = 0
                    last_skip = 2
                    intensity = 0.0
                    mono_int = 0.0
                    iso_distro = 1
                    sliced_mz = j[k]['m/z array']
                    sliced_int = j[k]['intensity array']
                    charge_range = range(1, abs(max_charges)*2)
                    target_mz = glycan_info['Adducts_mz'][i]
                    second_peak = (glycan_info['Isotopic_Distribution_Masses'][1]+adduct_mass)/adduct_charge
                    sec_peak_rel_int = glycan_info['Isotopic_Distribution'][1]
                    last_peak = (glycan_info['Isotopic_Distribution_Masses'][-1]+adduct_mass)/adduct_charge
                    no_iso_peaks = len(glycan_info['Isotopic_Distribution_Masses'])
                    before_target = glycan_info['Adducts_mz'][i]-General_Functions.h_mass-General_Functions.tolerance_calc(tolerance[0], tolerance[1], glycan_info['Adducts_mz'][i])
                    mono_ppm = []
                    iso_actual = []
                    iso_target = []
                    bad_peaks_before_target = []
                    nearby_id = 0
                    checked_bad_peaks_before_target = False
                    iso_found = False
                    current_iso_peak1 = []
                    current_iso_peak2 = []
                    for l_l, l in enumerate(sliced_mz):
                        if iso_found and l > ((glycan_info['Isotopic_Distribution_Masses'][iso_distro]+adduct_mass)/adduct_charge) + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                            if len(current_iso_peak1) > 5: #here data is profile
                                iso_actual.append(max(current_iso_peak1))
                                iso_target.append(max(current_iso_peak2))
                            else:
                                iso_actual.append(sum(current_iso_peak1))
                                iso_target.append(sum(current_iso_peak2))
                            current_iso_peak1 = []
                            current_iso_peak2 = []
                            iso_distro += 1
                            iso_found = False
                        if not_good: #Here are checks for quick skips
                            break
                        if l_l == len(sliced_mz)-1 and not found:
                            if verbose:
                                verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                verbose_info.append("--------Couldn't find peak and array ends shortly afterwards.")
                            not_good = True
                            break
                        if target_mz > sliced_mz[-1]:
                            if verbose:
                                verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                verbose_info.append("--------Target mz outside acquired mz interval.")
                            not_good = True
                            break
                        if l < before_target: #This is the last check for quick skips
                            continue
                        if l > second_peak + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and iso_distro == 1: #This is the first check for quick skips that's dependent on mz array acquired data but independent of noise
                            if verbose:
                                verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                verbose_info.append("--------Couldn't check correct charge.")
                            not_good = True
                            break
                        if l > (target_mz + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l)) and not found:
                            if verbose:
                                verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                verbose_info.append("--------Couldn't find monoisotopic peak.")
                            not_good = True
                            break
                        if l > last_peak + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and found:
                            if verbose:
                                verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                verbose_info.append("--------Checked all available isotopologue peaks with no error.")
                                verbose_info.append("--------Isotopologue peaks found: "+str(iso_distro)+"/"+str(no_iso_peaks))
                            if iso_distro < min_isotops:
                                not_good = True
                            break
                        if l > target_mz + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and l < target_mz + General_Functions.h_mass and found:
                            for m in charge_range:
                                if m == 1:
                                    continue
                                expected_value = (l*m*0.0006)+0.1401 #based on linear regression of the relationship between masses and the second isotopic peak relative intensity of the average of different organic macromolecules
                                if m != adduct_charge and abs(l-(target_mz+(General_Functions.h_mass/m))) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and (sliced_int[l_l] > mono_int*expected_value*0.6):
                                    if verbose:
                                        verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                        verbose_info.append("--------Incorrect charge assigned. Peak intensity: "+str(sliced_int[l_l])+" Monoisotopic peak intensity: "+str(mono_int)+" Charge detected: "+str(m)+" Found/Mono: "+str(sliced_int[l_l]/mono_int))
                                    not_good = True
                                    break
                            if not_good:
                                break
                        if l > target_mz - General_Functions.h_mass - General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and l < target_mz - General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                            nearby_id = l_l
                            for m in charge_range:
                                if abs(l-(target_mz-(General_Functions.h_mass/m))) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                                    bad_peaks_before_target.append((sliced_int[l_l], m))
                                    break
                        if l > target_mz + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and abs(l-((glycan_info['Isotopic_Distribution_Masses'][iso_distro]+adduct_mass)/adduct_charge)) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                            if sliced_int[l_l] > mono_int*glycan_info['Isotopic_Distribution'][iso_distro]:
                                intensity += mono_int*glycan_info['Isotopic_Distribution'][iso_distro]
                            if sliced_int[l_l] <= mono_int*glycan_info['Isotopic_Distribution'][iso_distro]:
                                intensity += sliced_int[l_l]
                            current_iso_peak1.append(sliced_int[l_l])
                            current_iso_peak2.append(mono_int*glycan_info['Isotopic_Distribution'][iso_distro])
                            iso_found = True
                            continue
                        if not checked_bad_peaks_before_target and l > target_mz + General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                            checked_bad_peaks_before_target = True
                            for m in bad_peaks_before_target:
                                expected_value = 1/((l*m[1]*0.0006)+0.1401) #based on linear regression of the relationship between masses and the second isotopic peak relative intensity of the average of different organic macromolecules
                                if (m[0] > mono_int*expected_value*0.6):
                                    if verbose:
                                        verbose_info.append("------m/z "+str(l)+", int "+str(sliced_int[l_l]))
                                        verbose_info.append("--------Not monoisotopic.")
                                    not_good = True
                                    break
                            if not_good:
                                break
                        if sliced_int[l_l] < General_Functions.local_noise_calc(noise[j_j][k_k], l, avg_noise[j_j]): #Everything from here is dependent on noise level (currently: monoisotopic peak detection only)
                            continue
                        if l >= target_mz - General_Functions.tolerance_calc(tolerance[0], tolerance[1], l) and abs(l-target_mz) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], l):
                            mono_ppm.append(General_Functions.calculate_ppm_diff(l, target_mz))
                            intensity += sliced_int[l_l]
                            mono_int += sliced_int[l_l]
                            found = True
                            continue
                    for l_l in range(nearby_id, len(sliced_mz)):
                        if sliced_mz[l_l] > target_mz+General_Functions.tolerance_calc(tolerance[0], tolerance[1], target_mz) or l_l == len(sliced_mz)-1:
                            raw_data[i][j_j][1].append(0.0)
                            break
                        if abs(sliced_mz[l_l] - target_mz) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], target_mz):
                            raw_data[i][j_j][1].append(sliced_int[l_l])
                            break
                if not_good:
                    ppm_info[i][j_j][-1] = inf
                    iso_fitting_quality[i][j_j][-1] = 0.0
                    data[i][j_j][1].append(0.0)
                    continue
                else:
                    ppm_info[i][j_j][-1] = mean(mono_ppm)
#                    print(j[k]['retentionTime'], iso_actual, iso_target)
                    if len(iso_actual) > 0:
                        weights = [1, 0.4, 0.1, 0.05, 0.02, 0.003, 0.0002, 0.00001, 0.00001, 0.000001]
                        temp_relation = []
                        for l in range(len(iso_actual)):
                            if iso_actual[l] >= iso_target[l]:
                                relation = iso_target[l]/iso_actual[l]
                                if relation >= 0.7:
                                    temp_relation.append(((relation)+9)/10)
                                if relation < 0.7 and relation >= 0.5:
                                    temp_relation.append(((relation)+6)/7)
                                if relation < 0.5 and relation >= 0.3:
                                    temp_relation.append(((relation)+3)/4)
                                if relation < 0.3:
                                    temp_relation.append(relation)
                            else:
                                relation = iso_actual[l]/iso_target[l]
                                if relation >= 0.7:
                                    temp_relation.append(((relation)+9)/10)
                                if relation < 0.70 and relation >= 0.5:
                                    temp_relation.append(((relation)+6)/7)
                                if relation < 0.5 and relation >= 0.3:
                                    temp_relation.append(((relation)+3)/4)
                                if relation < 0.3:
                                    temp_relation.append(relation)
                        R_sq = numpy.average(temp_relation, weights = weights[:len(temp_relation)])
                    if len(iso_actual) == 0:
                        R_sq = 0.0
                    iso_fitting_quality[i][j_j][-1] = R_sq
                    data[i][j_j][1].append(intensity)
    return data, ppm_info, iso_fitting_quality, verbose_info, raw_data
    
def eic_smoothing(y, lmbd = 100, d = 2):
    '''Implementation of the Whittaker smoothing algorithm,
    based on the work by Eilers [1].

    [1] P. H. C. Eilers, "A perfect smoother", Anal. Chem. 2003, (75), 3631-3636
    
    The larger 'lmbd', the smoother the data.
    For smoothing of a complete data series, sampled at equal intervals

    This implementation uses sparse matrices enabling high-speed processing
    of large input vectors
    
    Parameters
    ----------    
    y       : vector containing raw data
    lmbd    : parameter for the smoothing algorithm (roughness penalty)
    d       : order of the smoothing 

    Returns
    -------
    z : vector of the smoothed data.
    '''
    if y[0][len(y[0])//2]-y[0][(len(y[0])//2)-1] > 0.07: #this means that there are less than 15 data points per minute
        return y[0], y[1]
    array = numpy.array(y[1])
    m = len(array)
    E = sparse.eye(m, format='csc')
    D = General_Functions.speyediff(m, d, format='csc')
    coefmat = E + lmbd * D.conj().T.dot(D)
    z = splu(coefmat).solve(array)
    for i_i, i in enumerate(z):
        if i < 1:
            z[i_i] = 0.0
    return y[0], list(z)
    
def peak_curve_fit(rt_int, 
                   peak):
    '''Calculates the fitting between the actual peak and an ideal peak based
    on a calculated gaussian bell curve.
    
    Parameters
    ----------
    rt_int : list
        A list containing two lists: the first one has all the retention times, the
        second one has all the intensities.
        
    peak : dict
        A dictionary containing all sorts of identified peaks information.
        
    Uses
    ----
    General_Functions.normpdf : float
        Calculates the intensity of a gaussian bell curve at the x-axis point x
        with the set parameters.
        
    numpy.corrcoef : matrix
        Return Pearson product-moment correlation coefficients.
        
    Returns
    -------
    fits_list : tuple
        A tuple containing the R_sq of the curve fitting and plotting information
        of the actual and ideal peak curves.
    '''
    x = rt_int[0][peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1]
    y = rt_int[1][peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1]
    interval = x[len(x)//2]-x[(len(x)//2)-1]
    fits_list = []
    for j in range(int(0.2/interval)):
        before = []
        after = []
        y_adds = []
        for k in range(j):
            before.append(x[0]-k*interval)
            after.append(x[-1]+k*interval)
            y_adds.append(0.0)
        before = sorted(before)
        temp_x = before+x+after
        temp_y = y_adds+y+y_adds
        max_amp_id = temp_y.index(max(temp_y))
        counts_max = 0
        for i in temp_y:
            if i == temp_y[max_amp_id]:
                counts_max += 1
        max_amp_id += int(counts_max/2)
        y_gaussian = []
        for i in temp_x:
            y_gaussian.append(General_Functions.normpdf(i, temp_x[max_amp_id], (temp_x[-1]-temp_x[0])/6))
        scaler = (temp_y[max_amp_id]/y_gaussian[max_amp_id])
        y_gaussian_scaled = []
        for j in y_gaussian:
            y_gaussian_scaled.append(j*scaler)
        if len(temp_y[len(y_adds):len(temp_y)-len(y_adds)]) <= 2:
            temp_relation = []
            for l in range(len(temp_y[len(y_adds):len(temp_y)-len(y_adds)])):
                if temp_y[len(y_adds):len(temp_y)-len(y_adds)][l] >= y_gaussian_scaled[len(y_adds):len(temp_y)-len(y_adds)][l]:
                    temp_relation.append(temp_y[len(y_adds):len(temp_y)-len(y_adds)][l]/y_gaussian_scaled[len(y_adds):len(temp_y)-len(y_adds)][l])
                else:
                    temp_relation.append(y_gaussian_scaled[len(y_adds):len(temp_y)-len(y_adds)][l]/temp_y[len(y_adds):len(temp_y)-len(y_adds)][l])
            R_sq = mean(temp_relation)
        else:
            corr_matrix = numpy.corrcoef(temp_y[len(y_adds):len(temp_y)-len(y_adds)], y_gaussian_scaled[len(y_adds):len(temp_y)-len(y_adds)])
            corr = corr_matrix[0,1]
            R_sq = corr**2
        fits_list.append((R_sq, temp_x[len(y_adds):len(temp_y)-len(y_adds)], temp_y[len(y_adds):len(temp_y)-len(y_adds)], y_gaussian_scaled[len(y_adds):len(temp_y)-len(y_adds)]))
    fits_list = sorted(fits_list, reverse=True)
    return fits_list[0]
    
def iso_fit_score_calc(iso_fits,
                       peak):
    '''Calculates the mean isotopic fitting score of a given peak.
    
    Parameters
    ----------
    iso_fits : list
        A list of isotopic fitting scores calculated from eic_from_glycan.
        
    peak : dict
        A dictionary containing all sorts of identified peaks information.
        
    Uses
    ----
    statistics.mean : float
        Return the sample arithmetic mean of data which can be a sequence or iterable.
        
    Returns
    -------
    float
       The arithmetic mean of the isotopic fitting scores for the given peak.
    '''
    weights_list = []
    for i in range(peak['peak_interval_id'][1]-peak['peak_interval_id'][0]+1):
        weights_list.append(General_Functions.normpdf(i, (peak['peak_interval_id'][1]-peak['peak_interval_id'][0]+1)/2, (peak['peak_interval_id'][1]-peak['peak_interval_id'][0])/6))
    scaler = (1/max(weights_list))
    weights_list_scaled = []
    for i in weights_list:
        weights_list_scaled.append(i*scaler)
    regular_mean = numpy.average(iso_fits[peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1], weights = weights_list_scaled)
    return regular_mean
    
def average_ppm_calc(ppm_array,
                     tolerance,
                     peak):
    '''Calculates the arithmetic mean of the PPM differences of a given peak.
    
    Parameters
    ----------
    ppm_array : list
        A list containing ppm differences.
        
    tolerance : tuple
        First index contains the unit of the tolerance and the second one is the value of 
        that unit and the third index is the mz of the glycan.
        
    peak : dict
        A dictionary containing all sorts of identified peaks information.

    Uses
    ----
    General_Functions.calculate_ppm_diff : float
        Calculates the PPM difference between a mz and a target mz.
        
    Returns
    -------
    tuple
        A tuple containing the mean ppm differences and the number of missing points.
        The number of missing points in the peak has their ppm difference set to the
        tolerance of the analysis.
    '''
    ppms = []
    ppm_default = General_Functions.calculate_ppm_diff(tolerance[2]-General_Functions.tolerance_calc(tolerance[0], tolerance[1], tolerance[2]), tolerance[2])
    missing_points = 0
    weights_list = []
    for i in range(peak['peak_interval_id'][1]-peak['peak_interval_id'][0]+1):
        weights_list.append(General_Functions.normpdf(i, (peak['peak_interval_id'][1]-peak['peak_interval_id'][0]+1)/2, (peak['peak_interval_id'][1]-peak['peak_interval_id'][0])/6))
    scaler = (1/max(weights_list))
    weights_list_scaled = []
    for i in weights_list:
        weights_list_scaled.append(i*scaler)
    for i in ppm_array[peak['peak_interval_id'][0]:peak['peak_interval_id'][1]+1]:
        if i != inf:
            ppms.append(i)
        else:
            ppms.append(ppm_default)
            missing_points+= 1
    regular_mean = numpy.average(ppms, weights = weights_list_scaled)
    return regular_mean, missing_points

def peaks_from_eic(rt_int, 
                   rt_int_smoothed,
                   rt_interval,
                   min_ppp,
                   close_peaks,
                   glycan):
    '''Does multi peak-picking in a given smoothed EIC.
    
    Parameters
    ----------
    rt_int : list
        A list containing two lists: the first one has all the retention times, the
        second one has all the intensities.
        
    rt_int_smoothed : list
        A list containing two lists: the first one has all the retention times, the
        second one has all the SMOOTHED intensities.
        
    rt_interval : tuple
        A tuple where the first index contains the beggining time of the retention time
        interval you wish to analyze and the second contains the end time.
        
    min_ppp : tuple
        A tuple where the first index contains a boolean indicating whether or not to
        consider this parameter and the second one containing the minimum amount of
        data points per peak. If min_ppp[0] is set to False, calculates this automatically.
        
    close_peaks : tuple
        A tuple where the first index contains a boolean indicating whether or not to
        consider this parameter and the second one contains the amount of peaks to save.
        If close_peaks[0] is set to True, selects only the most intense peak and its 
        close_peaks[1] surrounding peaks.
        
    Returns
    -------
    peaks : list
        A list of dictionaries with each one containing numerous peaks information.
    '''
    peaks = []
    temp_start = 0
    temp_max = 0
    temp_max_id_iu = 0
    going_up = False
    going_down = False
    # counter = 0
    # max_counter = int(0.1/(rt_int[0][-1]-rt_int[0][-2]))
    slope_threshold = 30 #this number indicates the minimum angle to consider a peak
    for i_i, i in enumerate(rt_int_smoothed[1]):
        if (rt_int[0][i_i] >= rt_interval[1] or rt_int[0][i_i] == rt_int[0][-2]):
            break    
        if i > 0 and rt_int[0][i_i] >= rt_interval[0]:
            arc_tan_slope = atan(rt_int_smoothed[1][i_i+1]/i)*(180/pi) #arctan of slope
            if rt_int_smoothed[1][i_i+1] < i:
                arc_tan_slope = -arc_tan_slope
            if (going_up or going_down) and rt_int[1][i_i] > temp_max:
                temp_max = rt_int[1][i_i]
                temp_max_id_iu = i_i
            if (going_up and (arc_tan_slope < 0 or rt_int_smoothed[1][i_i] == 0)):
                # counter+=1
                # if counter <= max_counter:
                    # continue
                # elif counter > max_counter:
                    # counter = 0
                going_up = False
                going_down = True
            if (going_down and (arc_tan_slope >= 0 or rt_int_smoothed[1][i_i] == 0.0)):
                min_id = i_i
                if rt_int_smoothed[1][i_i+1] < i:
                    min_id = i_i+1
                temp_peak_width = (rt_int[0][min_id]-rt_int[0][temp_start])
                going_down = False
                good = False
                if min_ppp[0]:
                    if i_i-temp_start >= min_ppp[1] or glycan == "Internal Standard":
                        good = True
                else:
                    if min_id-temp_start >= int(0.2/(rt_int[0][-1]-rt_int[0][-2])) or glycan == "Internal Standard":
                        good = True
                if good:
                    peaks.append({'id': temp_max_id_iu, 'rt': rt_int[0][temp_max_id_iu], 'int': temp_max, 'peak_width': temp_peak_width, 'peak_interval': (rt_int[0][temp_start], rt_int[0][min_id]), 'peak_interval_id': (temp_start, min_id)})
                    temp_max = 0
                    temp_max_id_iu = 0
                if not good:
                    temp_max = 0
                    temp_max_id_iu = 0
            if (i_i != 0 and arc_tan_slope >= slope_threshold and not going_up and not going_down):
                temp_start = i_i-1
                going_up = True
    if close_peaks[0] or glycan == "Internal Standard":
        peaks = sorted(peaks, key=lambda x: x['int'], reverse = True)
        for i_i, i in enumerate(peaks):
            peaks[i_i]['proximity'] = abs(i['rt']-peaks[0]['rt'])
        peaks = sorted(peaks, key=lambda x: x['proximity'])
        if glycan == "Internal Standard":
            close_peaks = (True, 1)
        return sorted(peaks[:close_peaks[1]], key=lambda x: x['rt'])
    return sorted(peaks, key=lambda x: x['rt'])

def peaks_auc_from_eic(rt_int,
                       ms1_index,
                       peaks):
    '''Calculates the area under curve of the picked peaks based on the EIC given.
    Overlapped portions of peaks are calculated as fractional proportion, based on peaks
    max intensity.

    Parameters
    ----------
    rt_int : list
        A list containing two synchronized lists, the first one containing retention
        times and the second one containing intensity.
        
    ms1_indexes : dict
        A dictionary where each key is the ID of a file and each value is a list
        containing the indexes of all MS1 scans in the file.

    peaks : list
        A list of dictionaries with each one containing numerous peaks information.

    Returns
    -------
    auc : list
        A list of AUCs, with each index containing a float of the AUC of a peak.
    '''
    auc = []
    for i in peaks:
        temp_auc = 0
        for j in range(i['peak_interval_id'][0], i['peak_interval_id'][1]):
            temp_auc+=rt_int[1][j]
        auc.append(temp_auc)
    return auc



[library_building]
use_custom_glycans_list = no
custom_glycans_list = H3N2, H5N2, H5N4S2F1
; Allows you to only search for target glycans by inputing them in the "custom_glycans_list", as formulas. If True, overrides imp_library. Monosaccharides accepted: Hexoses (H), HexNAc (N), Acetyl Sialic Acid (S), Glycolyl Sialic Acid (G), Deoxyhexose (F). Case sensitive.
min_monos = 5
max_monos = 18
min_hex = 3
max_hex = 10
min_hexnac = 2
max_hexnac = 8
min_sia = 0
max_sia = 4
min_fuc = 0
max_fuc = 2
min_ac = 0
max_ac = 4
min_gc = 0
max_gc = 0
; Specify the minimum and maximum monosaccharides amounts for generating a library.
force_nglycan = yes
; Used to force some monosaccharides compositions associated with N-Glycans biologically known features.
max_adducts = H3Na1
; Indicates the desired adducts and their maximum amount. H3Na1 means a maximum of 3 Hydrogens and a maximum of 1 Sodium per adduct combination. Case sensitive.
max_charges = 3
; Limits the maximum amount of calculated charges for each glycan. Set to a negative value if you want to do negative mode analysis.
tag_mass = 133.0644
; If a reducing end tag is added to the glycans, insert its added mass here. If no reducing end tag is added to the glycans, set this value to 0. Procainamide: 219.1735; Girard Reagent P: 133.0644 (deprotonated, neutral)
permethylated = no
; If the sample was permethylated, set this parameter to "yes". Doesn't take into account partial permethylations.
reduced = no
; If the sample doesn't have a tag and the glycans had their reducing end reduced, set this to "yes".
fast_iso = yes
; Allows you to calculate the isotopic distribution of glycans based only on carbon isotopes (fast, innacurate) or on all the atoms isotopes (VERY SLOW, very accurate). If not used, library building may take many hours.
high_resolution_isotopic_dist = no
; If not used, doesn't clump isotope peaks together, meaning that you'll have more than one isotopic peak in a 1 Da interval. Only use this if you have fast_iso on. Useful for when analyzing very high resolution data, such as data acquired on a FT mass spectrometers.
internal_standard_mass = 0.0
; If using an internal standard, insert its mass here for the script to calculate its area.
imp_library = no
; Can be used to import a glycans library previously generated by this script. Glycan library must be in working directory (set below, on analysis_parameters).
exp_library = no
; Export the library you generate to use in future analysis without having to build a new library. Also creates an excel file containing a human-readable version of the library generated.
only_gen_lib = no
; Determines if the script should stop after generating library or proceed with data analysis. 

[analysis_parameters]
multithreaded_analysis = no
threads_number = 20
; Allows you to split the execution and generated library into multiple pieces for multithreaded execution. After running GlycoGenius, there will be several .py files in working directory. Run the ones named Multithreaded_0.py-Multithreaded_n.py in whatever time you want, in how many amounts at a time you want.
analyze_ms2 = yes
force_fragments_to_glycans = yes
unrestricted_fragments = no
reporter_ions = N1T1, 366.14
ret_time_tolerance_ms2 = 0.2
; Allows to analyze ms2 data, as well. Fragments identified will be associated with each glycan. You can choose to filter identified fragments by monosaccharides compositions, in order to avoid reporting fragments that aren't compatible with detected precursor. If unrestricted_fragments is used, it searches for glycans in every ms2 scan, regardless if the glycan was found in full scan. This will take a bit longer. If you have any reporter ions set, the program will only display fragments of MS2 spectra that contains said reporter ions. Can be based on glycans formula (with T being the reducing end of the glycan, including possibly the tag, if used) or an mz. reporter_ions can be changed on raw data reanalysis. ret_time_tolerance_ms2 indicates the retention time tolerance for attributing a given MS2 spectra to a specific peak.
accuracy_unit = mz
; Determines the units of mz tolerance to be used by the script. Options: 'ppm' or 'pw'. 'ppm' = Particles per Million, where 10 ppm is around 0.01 mz tolerance at mz 1000, 'mz' = Fixed mz tolerance from centroid, 0.01 mz means it tolerates a 0.01 variance in mz
accuracy_value = 0.01
; The value for the accuracy_unit parameter. You can use a broader accuracy value and then filter raw data using max_ppm, but this may lead to false positives.
ret_time_begin = 1
ret_time_end = 80
ret_time_tolerance = 0.2
; The minimum and maximum retention time, in minutes, used for various portions of the script. A shorter interval of ret_time makes the script run faster, so try to trim your sample as much as possible, if you know when your analytes are leaving the column. Set the retention time tolerance used for and same peak identification. ret_time_tolerance can be reapplied on raw data reanalysis.
custom_min_points_per_peak = no
number_points_per_peak = 5
; If used, set the minimum number of datapoints to consider a chromatogram peak part of the raw dataset. If left on False it calculates automatically.
limit_peaks_picked = no
max_number_peaks = 5
; If used, picks only the most intense peak on the EIC and up to [max_number_peaks]-1 other peaks closest to it. Warning: This may reduce the range of your results.
max_ppm = 10
; Maximum PPM for data curation. If value is greater than equivalent accuracy_value, data won't be filtered by this criteria, as it was already filtered during processing by accuracy_value. Can be reapplied on raw data reanalysis.
isotopic_fitting_score = 0.95
; Minimum score of the isotopic distribution fitting in order to consider a mz peak viable. Can be reapplied on raw data reanalysis.
curve_fitting_score = 0.95
; Minimum score for the chromatogram peak curve fitting to a gaussian to consider a viable peak. Can be reapplied on raw data reanalysis.
signal_to_noise = 3
; Minimum signal-to-noise ratio to consider a chromatogram peak viable. Can be reapplied on raw data reanalysis.
custom_noise_level = no
noise_levels = 0, 0
; Sets a custom level of noise for each sample, ignoring the automatic noise calculation. Noise for each sample is comma separated in "noise_levels". Warning: NOT RECOMMENDED unless you're really sure about what you're doing.
samples_path =
; A comma separated list of full path to files to be analyzed together.
working_path =
; Directory to load and save files from script.
plot_metaboanalyst = no
metaboanalyst_groups = CONTROL, TREATED
; Here you set up whether or not you want to output a .csv file to be used for plotting data using metaboanalyst. If you want that, you must specify your sample groups, comma separated. Sample groups specified must be present in sample filenames for proper identification. If none is set, samples are defaulted to "ungrouped". Case sensitive. 
reanalysis = no
output_plot_data = no
; Reanalyzes raw data with new max_ppm, isotopic_fitting_score, curve_fitting_score and signal_to_noise criteria. Overrides any other setting besides these mentioned. First parameter  produces a new Results file, second parameter also produces a new Plotting Data file (in case you deleted your original one. The data in it will not be any different than the former one). Warning: If setting a stricter max_ppm criteria on reanalysis without remaking the whole execution with a new accuracy_value, data may still contain false positives.


# GlycoGenius: Glycomics Data Analysis Tool
# Copyright (C) 2023 by Hector Franco Loponte
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or 
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. It is accessible within the program files
# or by typing 'license' after running it stand-alone in the terminal
# by typing 'glycogenius'. If not, see <https://www.gnu.org/licenses/>.

import pathlib
import importlib
import_path = str(pathlib.Path(__file__).parent.resolve())
for i_i, i in enumerate(import_path):
    if i == "\\":
        import_path = import_path[:i_i]+"/"+import_path[i_i+1:]
        
#Absolute import of General_Functions
spec1 = importlib.util.spec_from_file_location("General_Functions", import_path+'/General_Functions.py')
General_Functions = importlib.util.module_from_spec(spec1)
spec1.loader.exec_module(General_Functions)

from pyteomics import mzxml, mzml, mass, auxiliary
from itertools import combinations_with_replacement
from re import split
from math import inf
import sys
import datetime

##---------------------------------------------------------------------------------------
##Library generating-associated functions (these functions make vast use of the general
## functions).

def generate_glycans_library(min_max_mono,
                             min_max_hex,
                             min_max_hexnac,
                             min_max_sialics,
                             min_max_fuc,
                             min_max_ac,
                             min_max_gc,
                             n_glycan):
    '''Generates a list of combinatorial analysis of monosaccharides from the minimum
    amount of monosaccharides to the maximum amount of monosaccharides set, then trims it
    based on the specific monosaccharides criteria inserted.

    Parameters
    ----------
    min_max_mono : tuple
        Minimum and maximum amount of monosaccharides for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_hex : tuple
        Minimum and maximum amount of hexoses for the hypotethical glycans in the library.
        ie. (5, 20).

    min_max_hexnac : tuple
        Minimum and maximum amount of N-Acetyl hexosamines for the hypotethical glycans
        in the library. ie. (5, 20).

    min_max_sialics : tuple
        Minimum and maximum amount of sialic acids for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_fuc : tuple
        Minimum and maximum amount of deoxyhexoses for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_ac : tuple
        Minimum and maximum amount of N-Acetyl Neuraminic acids for the hypotethical
        glycans in the library. ie. (5, 20).

    min_max_gc : tuple
        Minimum and maximum amount of N-Glicolyl Neuraminic acids for the hypotethical
        glycans in the library. ie. (5, 20).

    n_glycan : boolean
        Indicates whether the function should force strict conditions based on the
        biological knowledge of glycans in order to avoid possible false positives when
        analysing N-glycans.

    Uses
    ----
    General_Functions.count_seq_letters(string) : string
        If you make anything with itertools for combinatorial analysis, it will produce a
        string that's not very human readable. This converts it into a human readable
        form.

    itertools.combinations_with_replacement : generator
        Return r length subsequences of elements from the input iterable allowing
        individual elements to be repeated more than once.

    General_Functions.sum_monos(*compositions) : dict
        Sums the monosaccharides of two glycan compositions.

    Returns
    -------
    glycans : list
        A list containing dictionaries of the monosaccharides compositions of all the
        glycans generated.
    '''
    glycans = []
    def_glycan_comp = {"H": 0, "N": 0, "S": 0, "F": 0, "G": 0}
    for i in range(min_max_mono[0], min_max_mono[1]+1):
        for j in combinations_with_replacement("HNSFG", i):
            glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
    to_be_removed = []
    for i_i, i in enumerate(glycans):
        if ((i['H'] < min_max_hex[0]) or (i['H'] > min_max_hex[1])
            or (i['N'] < min_max_hexnac[0]) or (i['N'] > min_max_hexnac[1])
            or (i['S']+i['G'] < min_max_sialics[0])
            or (i['S']+i['G'] > min_max_sialics[1])
            or (i['F'] < min_max_fuc[0]) or (i['F'] > min_max_fuc[1])
            or (i['S'] < min_max_ac[0]) or (i['S'] > min_max_ac[1])
            or (i['G'] < min_max_gc[0]) or (i['G'] > min_max_gc[1])):
            to_be_removed.append(i)
    if n_glycan:
        for i_i, i in enumerate(glycans):
            if ((i['S']+i['G'] > i['N']-2)
                or (i['F'] >= i['N']) or (i['S']+i['G'] > i['H']-3)
                or (i['H'] < 3) or (i['N'] < 2)):
                if i not in to_be_removed:
                    to_be_removed.append(i)
    for i in to_be_removed:
        glycans.remove(i)
    return glycans

def full_glycans_library(library,
                         max_adducts,
                         max_charges,
                         tag_mass = 0,
                         fast = True,
                         high_res = False,
                         internal_standard = 0.0,
                         permethylated = False,
                         reduced = False):
    '''Uses the generated glycans library and increments it with calculations of its
    mass, neutral mass with tag mass, its isotopic distribution pattern (intensoids) and
    the mz of the chosen adducts combinations.

    Parameters
    ----------
    library : list
        A list of dictionaries, each dict containing the composition of a hypothetical
        glycan, as generated by generate_glycans_library.

    max_adducts : dict
        A dictionary with keys containing each possible atomic adducts (ie. 'H', 'Na',
        'K', etc.) and the maximum amount of such adducts as the values.

    max_charges : int
        The maximum amount of charges to calculate per glycan.
        
    tolerance : tuple
        First index contains the unit of the tolerance and the second one is the value of 
        that unit.

    tag_mass : float
        The tag's added mass to the glycans, if the glycans are tagged.
        Default = 0 (No Tag).

    fast : boolean
        Makes the isotopic distribution calculation fast (less accurate) or not (more
        accurate).
        Default = True
        
    high_res : boolean
        Decides whether to clump (if set to False) or not (if set to True) the neighbouring
        isotopic envelope peaks. Only works if fast is set to False.
        Default = False
        
    permethylated : boolean
        Whether or not the sample was permethylated.
        
    reduced : boolean
        Whether or not the sample was reduced.

    Uses
    ----
    General_Functions.calculate_comp_from_mass(tag_mass) : dict
        Calculates the composition of a molecule based on its mass. Intended to use with
        small tags added to the glycans.
        
    General_Functions.gen_adducts_combo(adducts, max_charge) : list
        Generates a list of dictionaries with compositions of adducts combinations,
        based on parameters set.

    General_Functions.comp_to_formula(composition) : string
        Transforms a composition dictionary into string formula.

    General_Functions.sum_atoms(*compositions) : dict
        Sums the atoms of two compositions.

    General_Functions.glycan_to_atoms(glycan_composition) : dict
        Calculates the amounts of atoms based on glycan monosaccharides.

    General_Functions.form_to_comp(string) : dict
        Separates a molecular formula or monosaccharides composition of glycans into a
        dictionary with each atom/monosaccharide as a key and its amount as value.

    pyteomics.mass.calculate_mass(*args, **kwargs) : float
        Calculates the monoisotopic mass of a polypeptide defined by a sequence string,
        parsed sequence, chemical formula or Composition object.

    General_Functions.calculate_isotopic_pattern(glycan_atoms,
                               tolerance,
                               fast=True) : list
        Calculates up to 5 isotopic pattern peaks relative abundance in relation with
        the monoisotopic one.

    Returns
    -------
    full_library : dict
        A dictionary with each key containing the glycan formula and each key containing
        a dictionary with monosaccharides composition, atoms composition with tag,
        neutral mass, neutral mass with tag, isotopic distribution and the mzs of the
        glycans with the desired adducts combination.
    '''
    full_library = {}
    if tag_mass != 0:
        tag = General_Functions.calculate_comp_from_mass(tag_mass)
    else:
        tag = ({"C": 0, "O": 0, "N": 0, "H": 0}, 0.0)
    adducts_combo = General_Functions.gen_adducts_combo(max_adducts, max_charges)
    for i in library:
        i_formula = General_Functions.comp_to_formula(i)
        i_atoms = General_Functions.sum_atoms(General_Functions.glycan_to_atoms(i, permethylated), General_Functions.form_to_comp('H2O'))
        if tag[1] == 0.0:
            if permethylated:
                i_atoms = General_Functions.sum_atoms(i_atoms, {'C': 2, 'H': 4})
                if reduced:
                    i_atoms = General_Functions.sum_atoms(i_atoms, {'O': 1})
            if not permethylated and reduced:
                i_atoms = General_Functions.sum_atoms(i_atoms, {'H': 2})
        i_atoms_tag = General_Functions.sum_atoms(i_atoms, tag[0])
        i_neutral_mass = mass.calculate_mass(composition=i_atoms)
        i_neutral_tag = i_neutral_mass+tag[1]
        i_iso_dist = General_Functions.calculate_isotopic_pattern(i_atoms_tag, fast, high_res)
        iso_corrected = i_iso_dist[0]
        if fast:
            iso_corrected = []
            for j_j, j in enumerate(i_iso_dist[0]):
                if j_j == 1:
                    iso_corrected.append(abs(j*1.03))
                    continue
                if j_j == 2:
                    iso_corrected.append(abs(j*(1.5-(0.00055*sum(i_atoms_tag.values())))))
                    continue
                if j_j == 3:
                    iso_corrected.append(abs(j*(2.43-(0.00172*sum(i_atoms_tag.values())))))
                    continue
                if j_j == 4:
                    iso_corrected.append(abs(j*(5.5-(0.0062*sum(i_atoms_tag.values())))))
                    continue
                if j_j == 5:
                    iso_corrected.append(abs(j*(3.16-(0.00585*sum(i_atoms_tag.values())))))
                    continue
                else:
                    iso_corrected.append(j)
                    continue
        full_library[i_formula] = {}
        full_library[i_formula]['Monos_Composition'] = i
        full_library[i_formula]['Atoms_Glycan+Tag'] = i_atoms_tag
        full_library[i_formula]['Neutral_Mass'] = i_neutral_mass
        full_library[i_formula]['Neutral_Mass+Tag'] = i_neutral_tag
        full_library[i_formula]['Isotopic_Distribution'] = iso_corrected
        full_library[i_formula]['Isotopic_Distribution_Masses'] = i_iso_dist[1]
        full_library[i_formula]['Adducts_mz'] = {}
        for j in adducts_combo:
            charges = sum(j.values())
            mz = mass.calculate_mass(i_atoms_tag,
                                     charge = charges,
                                     charge_carrier = j,
                                     carrier_charge = charges)
            full_library[i_formula]['Adducts_mz'][General_Functions.comp_to_formula(j)] = mz
    if internal_standard != 0.0:
        i_formula = 'Internal Standard'
        i_neutral_mass = internal_standard
        if tag[1] == 0.0:
            if permethylated:
                i_neutral_mass = i_neutral_mass+mass.calculate_mass({'C': 2, 'H': 4})
                if reduced:
                    i_neutral_mass = i_neutral_mass+mass.calculate_mass({'O': 1})
            if not permethylated and reduced:
                i_neutral_mass = i_neutral_mass+mass.calculate_mass({'H': 2})
        else:
            i_neutral_mass = i_neutral_mass+tag_mass
        i_iso_dist = [[1.0, 0.9, 0.7, 0.4, 0.1], []]
        for i in range(len(i_iso_dist[0])):
            i_iso_dist[1].append(internal_standard+(i*General_Functions.h_mass))
        full_library[i_formula] = {}
        full_library[i_formula]['Monos_Composition'] = {"H": 0, "N": 0, "S": 0, "F": 0, "G": 0}
        full_library[i_formula]['Neutral_Mass'] = i_neutral_mass
        full_library[i_formula]['Neutral_Mass+Tag'] = i_neutral_mass
        full_library[i_formula]['Isotopic_Distribution'] = i_iso_dist[0]
        full_library[i_formula]['Isotopic_Distribution_Masses'] = i_iso_dist[1]
        full_library[i_formula]['Adducts_mz'] = {}
        for j in adducts_combo:
            charges = sum(j.values())
            mz = (i_neutral_mass+mass.calculate_mass(j))/charges
            full_library[i_formula]['Adducts_mz'][General_Functions.comp_to_formula(j)] = mz
    return full_library

def fragments_library(min_max_mono,
                      min_max_hex,
                      min_max_hexnac,
                      min_max_sialics,
                      min_max_fuc,
                      min_max_ac,
                      min_max_gc,
                      max_charges,
                      tolerance,
                      tag_mass,
                      permethylated,
                      reduced,
                      nglycan):
    '''Generates a list of combinatorial analysis of monosaccharides from the minimum
    amount of monosaccharides to the maximum amount of monosaccharides set, then uses 
    the generated library and increments it with calculations with a series of information,
    clumps the duplicated mzs together as one query and then the script can use this
    library as a fragments library for MS2 analysis.

    Parameters
    ----------
    min_max_mono : tuple
        Minimum and maximum amount of monosaccharides for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_hex : tuple
        Minimum and maximum amount of hexoses for the hypotethical glycans in the library.
        ie. (5, 20).

    min_max_hexnac : tuple
        Minimum and maximum amount of N-Acetyl hexosamines for the hypotethical glycans
        in the library. ie. (5, 20).

    min_max_sialics : tuple
        Minimum and maximum amount of sialic acids for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_fuc : tuple
        Minimum and maximum amount of deoxyhexoses for the hypotethical glycans in the
        library. ie. (5, 20).

    min_max_ac : tuple
        Minimum and maximum amount of N-Acetyl Neuraminic acids for the hypotethical
        glycans in the library. ie. (5, 20).

    min_max_gc : tuple
        Minimum and maximum amount of N-Glicolyl Neuraminic acids for the hypotethical
        glycans in the library. ie. (5, 20).

    max_charges : int
        The maximum amount of charges to calculate per glycan.

    tolerance : float
        The mz acceptable tolerance. Used to calculate the intensoids, at this point.

    tag_mass : float
        The tag's added mass to the glycans, if the glycans are tagged.
        Default = 0 (No Tag).
        
    permethylated : boolean
        Whether the glycan is permethylated or not.
        
    reduced : boolean
        Whether the glycan is reduced or not.

    Uses
    ----
    General_Functions.count_seq_letters(string) : string
        If you make anything with itertools for combinatorial analysis, it will produce a
        string that's not very human readable. This converts it into a human readable
        form.

    itertools.combinations_with_replacement : generator
        Return r length subsequences of elements from the input iterable allowing
        individual elements to be repeated more than once.

    General_Functions.sum_monos(*compositions) : dict
        Sums the monosaccharides of two glycan compositions.
        
    General_Functions.calculate_comp_from_mass(tag_mass) : dict
        Calculates the composition of a molecule based on its mass. Intended to use with
        small tags added to the glycans.
        
    General_Functions.gen_adducts_combo(adducts, max_charge) : list
        Generates a list of dictionaries with compositions of adducts combinations,
        based on parameters set.

    General_Functions.comp_to_formula(composition) : string
        Transforms a composition dictionary into string formula.

    General_Functions.sum_atoms(*compositions) : dict
        Sums the atoms of two compositions.

    General_Functions.glycan_to_atoms(glycan_composition) : dict
        Calculates the amounts of atoms based on glycan monosaccharides.

    pyteomics.mass.calculate_mass(*args, **kwargs) : float
        Calculates the monoisotopic mass of a polypeptide defined by a sequence string,
        parsed sequence, chemical formula or Composition object.

    Returns
    -------
    glycans : list
        A list containing dictionaries with informations about each fragment generated.
    '''
    print("Building fragments library...", end = "", flush = True)
    glycans = []
    def_glycan_comp = {"H": 0, "N": 0, "S": 0, "F": 0, "G": 0, "T" : 0}
    for i in range(1, 18):
        for j in combinations_with_replacement("HNSFGT", i):
            glycans.append(General_Functions.sum_monos(def_glycan_comp, General_Functions.count_seq_letters("".join(j))))
    to_be_removed = []
    for i_i, i in enumerate(glycans):
        if ((i['T'] > 1) 
            or (i['T'] == 1 and i['N'] == 0)
            or (i['H'] > min_max_hex[1])
            or (i['N'] > min_max_hexnac[1])
            or (i['S']+i['G'] > min_max_sialics[1])
            or (i['F'] > min_max_fuc[1])
            or (i['S'] > min_max_ac[1])
            or (i['G'] > min_max_gc[1])):
            to_be_removed.append(i_i)
        if nglycan and i_i not in to_be_removed: #some rules and hardcoded exceptions for N-Glycans
            if ((i['T'] == 1 and sum(i.values()) < 8 and i['S']+i['G'] > 0)
                or (sum(i.values()) < 6 and i['S'] >= 1 and i['N'] > 1)
                or (i['H'] > 0 and i['T'] == 1 and i['N'] < 2)
                or (i['H'] == 2 and i['N'] == 1 and i['S'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
                or (i['H'] == 0 and i['N'] == 1 and (i['S'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
                or (i['H'] == 1 and i['N'] == 3 and i['S'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
                or (i['H'] > 1 and i['N'] == 0 and (i['S'] == 1 or i['G'] == 1) and i['F'] == 0 and i['T'] == 0)
                or (i['H'] == 1 and i['N'] == 1 and i['S'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)
                or (i['H'] == 3 and i['N'] == 1 and i['S'] == 0 and i['F'] == 0 and i['G'] == 0 and i['T'] == 1)):
                to_be_removed.append(i_i)
    for i in sorted(to_be_removed, reverse = True):
        del glycans[i]
    if tag_mass != 0:
        tag = General_Functions.calculate_comp_from_mass(tag_mass)
    else:
        tag = ({"C": 0, "O": 0, "N": 0, "H": 0}, 0.0)
    adducts_combo = General_Functions.gen_adducts_combo({'H' : 2}, max_charges)
    frag_library = []
    combo_frags_lib = []
    adducts_mz = [[], [], []]
    for i_i, i in enumerate(glycans):
        for j_j, j in enumerate(range(-1, 2)):
            if j < 0:
                i_formula = General_Functions.comp_to_formula(i)+str(j)+'H2O'
            elif j > 0:
                i_formula = General_Functions.comp_to_formula(i)+'+'+str(j)+'H2O'
            else:
                i_formula = General_Functions.comp_to_formula(i)
            glycan_atoms = General_Functions.glycan_to_atoms(i, permethylated)
            glycan_atoms['H'] += j*2
            glycan_atoms['O'] += j*1
            i_atoms = glycan_atoms
            if tag[1] == 0.0:
                if permethylated:
                    i_atoms = General_Functions.sum_atoms(i_atoms, {'C': 2, 'H': 4})
                    if reduced:
                        i_atoms = General_Functions.sum_atoms(i_atoms, {'O': 1})
                if not permethylated and reduced:
                    i_atoms = General_Functions.sum_atoms(i_atoms, {'H': 2})
            i_neutral_mass = mass.calculate_mass(composition=i_atoms)
            if i['T'] == 1:
                i_atoms_tag = General_Functions.sum_atoms(i_atoms, tag[0])
                i_neutral_tag = i_neutral_mass+tag[1]
                if tag[1] == 0.0:
                    if permethylated:
                        i_atoms = General_Functions.sum_atoms(i_atoms, {'C': 1, 'H': 2})
                        if reduced:
                            i_atoms = General_Functions.sum_atoms(i_atoms, {'O': 1})
                    if not permethylated and reduced:
                        i_atoms = General_Functions.sum_atoms(i_atoms, {'H': 2})
            else:
                i_atoms_tag = i_atoms
                i_neutral_tag = i_neutral_mass
            frag_library.append({})
            index = (i_i*3)+j_j
            frag_library[index]['Formula'] = i_formula
            frag_library[index]['Monos_Composition'] = i
            frag_library[index]['Adducts_mz'] = {}
            if tag[1] == 0.0: #this skips and marks for removal the reducing end without water
                if j < 0:
                    continue
            for j in adducts_combo:
                charges = sum(j.values())
                mz = mass.calculate_mass(i_atoms_tag,
                                         charge = charges,
                                         charge_carrier = j,
                                         carrier_charge = charges)
                found = False
                for k_k, k in enumerate(adducts_mz[0]):
                    if abs(mz - k) <= General_Functions.tolerance_calc(tolerance[0], tolerance[1], mz):
                        combo_frags_lib.append({})
                        combo_frags_lib[-1]['Formula'] = str(frag_library[adducts_mz[2][k_k]]['Formula']+'_'+adducts_mz[1][k_k]+'/'+frag_library[index]['Formula']+'_'+General_Functions.comp_to_formula(j))
                        combo_frags_lib[-1]['Adducts_mz'] = {}
                        combo_frags_lib[-1]['Adducts_mz'][General_Functions.comp_to_formula(j)] = mz
                        combo_frags_lib[-1]['Adducts_mz'][adducts_mz[1][k_k]] = mz
                        try:
                            del frag_library[adducts_mz[2][k_k]]['Adducts_mz'][adducts_mz[1][k_k]]
                        except:
                            pass
                        adducts_mz[0].append(mz)
                        adducts_mz[1].append(General_Functions.comp_to_formula(j))
                        adducts_mz[2].append(index)
                        found = True
                        break
                if not found:
                    adducts_mz[0].append(mz)
                    adducts_mz[1].append(General_Functions.comp_to_formula(j))
                    adducts_mz[2].append(index)
                    frag_library[index]['Adducts_mz'][General_Functions.comp_to_formula(j)] = mz
    to_remove = []
    for i_i, i in enumerate(frag_library):
        if len(i) == 0 or len(i['Adducts_mz']) == 0:
            to_remove.append(i_i)
    for i_i in sorted(to_remove, reverse = True):
        del frag_library[i_i]
    frag_library = frag_library+combo_frags_lib
    print("Done!")
    return frag_library


# GlycoGenius: Glycomics Data Analysis Tool
# Copyright (C) 2023 by Hector Franco Loponte
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or 
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. It is accessible within the program files
# or by typing 'license' after running it stand-alone in the terminal
# by typing 'glycogenius'. If not, see <https://www.gnu.org/licenses/>.

from . import Execution_Functions
from . import General_Functions
from pathlib import Path
import sys
import os
import datetime
import configparser

#-----------------------------------------------------------------------------

def main():
    custom_glycans_list = (False, [])
    min_max_monos = (0, 0)
    min_max_hex = (0, 0)
    min_max_hexnac = (0, 0)
    min_max_sia = (0, 0)
    min_max_fuc = (0, 0)
    min_max_ac = (0, 0)
    min_max_gc = (0, 0)
    force_nglycan = False
    max_adducts = {}
    max_charges = 0
    tag_mass = 0.0
    permethylated = False
    reduced = False
    fast_iso = True
    high_res = False
    internal_standard = 0.0
    imp_exp_library = (False, False)
    only_gen_lib = False

    multithreaded_analysis = (False, 1)
    analyze_ms2 = (False, False, False)
    reporter_ions = []
    accuracy_unit = 'mz'
    accuracy_value = 0.01
    ret_time_interval = (0, 99999, 0.2)
    rt_tolerance_frag = 0.2
    min_isotopologue_peaks = 2
    min_ppp = (False, 0)
    close_peaks = (False, 3)
    max_ppm = 10
    iso_fit_score = 0.9
    curve_fit_score = 0.9
    s_to_n = 3
    custom_noise = (False, [])
    samples_path = ''
    save_path = ''
    plot_metaboanalyst = (False, [])
    reanalysis = (False, False)

    multithreaded_execution = (False, 0, 0) #editted by multithreaded 1
    sneakpeek = (False, 0)
    verbose = False

    if not os.isatty(0) or multithreaded_execution[0]: #If multithreaded execution or pipelined parameters file
        Execution_Functions.print_header(False)
        config = configparser.ConfigParser()
        configs = ""
        for line in sys.stdin:              #editted by multithreaded 2
            configs+=line                   #editted by multithreaded 3
        config.read_string(configs)
        custom_glycans = config['library_building']['custom_glycans_list'].split(",")
        for i_i, i in enumerate(custom_glycans):
            custom_glycans[i_i] = i.strip()
            if len(i) == 0:
                custom_glycans = custom_glycans[:i_i]+custom_glycans[i_i+1:]
        custom_glycans_list = (config['library_building'].getboolean('use_custom_glycans_list'), custom_glycans)
        min_max_monos = (int(config['library_building']['min_monos']), int(config['library_building']['max_monos']))
        min_max_hex = (int(config['library_building']['min_hex']), int(config['library_building']['max_hex']))
        min_max_hexnac = (int(config['library_building']['min_hexnac']), int(config['library_building']['max_hexnac']))
        min_max_sia = (int(config['library_building']['min_sia']), int(config['library_building']['max_sia']))
        min_max_fuc = (int(config['library_building']['min_fuc']), int(config['library_building']['max_fuc']))
        min_max_ac = (int(config['library_building']['min_ac']), int(config['library_building']['max_ac']))
        min_max_gc = (int(config['library_building']['min_gc']), int(config['library_building']['max_gc']))
        force_nglycan = config['library_building'].getboolean('force_nglycan')
        max_adducts = General_Functions.form_to_comp(config['library_building']['max_adducts'])
        max_charges = int(config['library_building']['max_charges'])
        tag_mass = float(config['library_building']['tag_mass'])
        permethylated = config['library_building'].getboolean('permethylated')
        reduced = config['library_building'].getboolean('reduced')
        fast_iso = config['library_building'].getboolean('fast_iso')
        high_res = config['library_building'].getboolean('high_resolution_isotopic_dist')
        internal_standard = float(config['library_building']['internal_standard_mass'])
        imp_exp_library = (config['library_building'].getboolean('imp_library'), config['library_building'].getboolean('exp_library'))
        only_gen_lib = config['library_building'].getboolean('only_gen_lib')
        multithreaded_analysis = (config['analysis_parameters'].getboolean('multithreaded_analysis'), int(config['analysis_parameters']['threads_number']))
        analyze_ms2 = (config['analysis_parameters'].getboolean('analyze_ms2'), config['analysis_parameters'].getboolean('force_fragments_to_glycans'), config['analysis_parameters'].getboolean('unrestricted_fragments'))
        reporter_ions = config['analysis_parameters']['reporter_ions'].split(",")
        for i_i, i in enumerate(reporter_ions):
            reporter_ions[i_i] = i.strip()
            if len(i) == 0:
                reporter_ions = reporter_ions[:i_i]+reporter_ions[i_i+1:]
        rt_tolerance_frag = float(config['analysis_parameters']['ret_time_tolerance_ms2'])
        accuracy_unit = config['analysis_parameters']['accuracy_unit']
        accuracy_value = float(config['analysis_parameters']['accuracy_value'])
        ret_time_interval = (float(config['analysis_parameters']['ret_time_begin']), float(config['analysis_parameters']['ret_time_end']), float(config['analysis_parameters']['ret_time_tolerance']))
        min_ppp = (config['analysis_parameters'].getboolean('custom_min_points_per_peak'), int(config['analysis_parameters']['number_points_per_peak']))
        close_peaks = (config['analysis_parameters'].getboolean('limit_peaks_picked'), int(config['analysis_parameters']['max_number_peaks']))
        max_ppm = int(config['analysis_parameters']['max_ppm'])
        iso_fit_score = float(config['analysis_parameters']['isotopic_fitting_score'])
        curve_fit_score = float(config['analysis_parameters']['curve_fitting_score'])
        s_to_n = int(config['analysis_parameters']['signal_to_noise'])
        noise_levels = config['analysis_parameters']['noise_levels'].split(",")
        for i_i, i in enumerate(noise_levels):
            noise_levels[i_i] = int(i.strip())
        custom_noise = (config['analysis_parameters'].getboolean('custom_noise_level'), noise_levels)
        samples_path = config['analysis_parameters']['samples_path']
        samples_path = samples_path.strip()
        samples_path = samples_path.strip("'")
        samples_path = samples_path.strip("\"")
        for i_i, i in enumerate(samples_path):
            if i == "\\":
                samples_path = samples_path[:i_i]+"/"+samples_path[i_i+1:]
        if samples_path[-1] != "/":
            samples_path = samples_path+"/"
        save_path = config['analysis_parameters']['working_path']
        save_path = save_path.strip()
        save_path = save_path.strip("'")
        save_path = save_path.strip("\"")
        for i_i, i in enumerate(save_path):
            if i == "\\":
                save_path = save_path[:i_i]+"/"+save_path[i_i+1:]
        if save_path[-1] != "/":
            save_path = save_path+"/"
        metaboanalyst_groups = config['analysis_parameters']['metaboanalyst_groups'].split(",")
        for i_i in range(len(metaboanalyst_groups)-1, -1, -1):
            metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip()
            metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip("'")
            metaboanalyst_groups[i_i] = metaboanalyst_groups[i_i].strip("\"")
            if len(metaboanalyst_groups[i_i]) == 0:
                del metaboanalyst_groups[i_i]
        plot_metaboanalyst = (config['analysis_parameters'].getboolean('plot_metaboanalyst'), metaboanalyst_groups)
        reanalysis = (config['analysis_parameters'].getboolean('reanalysis'), config['analysis_parameters'].getboolean('output_plot_data'))
    else: #If no parameters file pipelines, run CLI
        parameters = Execution_Functions.interactive_terminal()
        Execution_Functions.print_sep()
        if parameters[0][0] == 69: #sneakpeek feature
            sneakpeek = (True, parameters[2])
            config = configparser.ConfigParser()
            configs = ""
            with open(parameters[1]+'glycogenius_parameters.ini', 'r') as f:
                for line in f:
                    configs+=line
            config.read_string(configs)
            max_ppm = int(config['analysis_parameters']['max_ppm'])
            iso_fit_score = float(config['analysis_parameters']['isotopic_fitting_score'])
            curve_fit_score = float(config['analysis_parameters']['curve_fitting_score'])
            s_to_n = int(config['analysis_parameters']['signal_to_noise'])
            analyze_ms2 = (config['analysis_parameters'].getboolean('analyze_ms2'), config['analysis_parameters'].getboolean('force_fragments_to_glycans'), config['analysis_parameters'].getboolean('unrestricted_fragments'))
            save_path = parameters[1]
            reanalysis = (True, True)
        if parameters[0][0] == 1 or parameters[0][0] == 2:
            if parameters[0][1] == 1:
                custom_glycans_list = (True, parameters[1])
                max_adducts = parameters[2]
                max_charges = parameters[3]
                tag_mass = parameters[4]
                fast_iso = parameters[5]
                high_res = parameters[6]
                permethylated = parameters[8]
                reduced = parameters[9]
            if parameters[0][1] == 2:
                min_max_monos = (parameters[1][0], parameters[1][1])
                min_max_hex = (parameters[1][2], parameters[1][3])
                min_max_hexnac = (parameters[1][4], parameters[1][5])
                min_max_fuc = (parameters[1][6], parameters[1][7])
                min_max_ac = (parameters[1][8], parameters[1][9])
                min_max_gc = (parameters[1][10], parameters[1][11])
                min_max_sia = (parameters[1][12], parameters[1][13])
                force_nglycan = parameters[1][14]
                max_adducts = parameters[2]
                max_charges = parameters[3]
                tag_mass = parameters[4]
                fast_iso = parameters[5]
                high_res = parameters[6]
                permethylated = parameters[8]
                reduced = parameters[9]
            if parameters[0][0] == 1:
                save_path = parameters[7]
                if save_path[-1] != "/":
                    save_path+= "/"
                only_gen_lib = True
            if parameters[0][0] == 2:
                analyze_ms2 = parameters[7]
                accuracy_unit = parameters[8]
                accuracy_value = parameters[9]
                ret_time_interval = (parameters[10][0], parameters[10][1], 0.2)
                min_isotopologue_peaks = parameters[11]
                max_ppm = parameters[12]
                iso_fit_score = parameters[13]
                curve_fit_score = parameters[14]
                s_to_n = parameters[15]
                samples_path = parameters[16]
                save_path = parameters[17]
                permethylated = parameters[18]
                reduced = parameters[19]
        if parameters[0][0] == 3:
            save_path = parameters[1]
            max_ppm = parameters[2]
            iso_fit_score = parameters[3]
            curve_fit_score = parameters[4]
            s_to_n = parameters[5]
            reanalysis = (True, True)
        if parameters[0][0] == 4:
            comments = parameters[1]
            save_path = parameters[2]
            Path(save_path).mkdir(exist_ok = True, parents = True)
            Execution_Functions.generate_cfg_file(save_path, comments)

    Path(save_path).mkdir(exist_ok = True, parents = True)
    if samples_path != '':
        Path(samples_path).mkdir(exist_ok = True, parents = True)
    samples_list = Execution_Functions.samples_path_to_list(samples_path)

#-----------------------------------------------------------------------------

    begin_time = datetime.datetime.now()

    if reanalysis[0]:
        Execution_Functions.output_filtered_data(curve_fit_score,
                                                 iso_fit_score,
                                                 s_to_n,
                                                 max_ppm,
                                                 reanalysis,
                                                 save_path,
                                                 multithreaded_analysis,
                                                 multithreaded_execution,
                                                 analyze_ms2[0],
                                                 analyze_ms2[2],
                                                 reporter_ions,
                                                 plot_metaboanalyst,
                                                 ret_time_interval[2],
                                                 rt_tolerance_frag,
                                                 sneakpeek)

    else:
        if multithreaded_execution[0]:
            print('Multithreaded Execution: '+str(multithreaded_execution[1]))
        samples_names = Execution_Functions.sample_names(samples_list)
        if not only_gen_lib and not reanalysis[0]:
            print("Sample files detected: "+str(len(samples_names)))
            for i in samples_names:
                print("--> "+i)
            Execution_Functions.print_sep()
        library = Execution_Functions.imp_exp_gen_library(multithreaded_analysis,
                                                          multithreaded_execution,
                                                          samples_names,
                                                          custom_glycans_list,
                                                          min_max_monos,
                                                          min_max_hex,
                                                          min_max_hexnac,
                                                          min_max_sia,
                                                          min_max_fuc,
                                                          min_max_ac,
                                                          min_max_gc,
                                                          force_nglycan,
                                                          max_adducts,
                                                          max_charges,
                                                          tag_mass,
                                                          fast_iso,
                                                          high_res,
                                                          imp_exp_library,
                                                          only_gen_lib,
                                                          save_path,
                                                          internal_standard,
                                                          permethylated,
                                                          reduced)
#line to add multithreaded library
        if multithreaded_execution[0]:
            library = full_library
        print('Library length: '+str(len(library)))
        Execution_Functions.print_sep()
        tolerance = (accuracy_unit, accuracy_value)
        print("Preparing files for analysis...", end = "", flush = True)
        data = Execution_Functions.list_of_data(samples_list)
        print("Done!")
        print("Indexing spectra...", end = "", flush = True)
        ms1_index = Execution_Functions.index_ms1_from_file(data)
        if analyze_ms2[0]:
            data = Execution_Functions.list_of_data(samples_list)
            ms2_index = Execution_Functions.index_ms2_from_file(data)
        print("Done!")
        lib_size = len(library)
        analyzed_data = Execution_Functions.analyze_files(library,
                                                          lib_size,
                                                          data,
                                                          ms1_index,
                                                          tolerance,
                                                          ret_time_interval,
                                                          min_isotopologue_peaks,
                                                          min_ppp,
                                                          max_charges,
                                                          custom_noise,
                                                          close_peaks,
                                                          fast_iso,
                                                          verbose)
        if analyze_ms2[0]:
            Execution_Functions.print_sep()
            analyzed_data = Execution_Functions.analyze_ms2(ms2_index, 
                                                            data, 
                                                            analyzed_data, 
                                                            ret_time_interval,
                                                            tolerance,
                                                            min_max_monos,
                                                            min_max_hex,
                                                            min_max_hexnac,
                                                            min_max_sia,
                                                            min_max_fuc,
                                                            min_max_ac,
                                                            min_max_gc,
                                                            max_charges,
                                                            tag_mass,
                                                            force_nglycan,
                                                            permethylated,
                                                            reduced,
                                                            analyze_ms2[1],
                                                            analyze_ms2[2],
                                                            ret_time_interval[2])
        Execution_Functions.print_sep()
        Execution_Functions.arrange_raw_data(analyzed_data,
                                             samples_names,
                                             multithreaded_analysis,
                                             multithreaded_execution,
                                             analyze_ms2[0],
                                             save_path)
        Execution_Functions.print_sep()
        Execution_Functions.output_filtered_data(curve_fit_score,
                                                 iso_fit_score,
                                                 s_to_n,
                                                 max_ppm,
                                                 reanalysis,
                                                 save_path,
                                                 multithreaded_analysis,
                                                 multithreaded_execution,
                                                 analyze_ms2[0],
                                                 analyze_ms2[2],
                                                 reporter_ions,
                                                 plot_metaboanalyst,
                                                 ret_time_interval[2],
                                                 rt_tolerance_frag,
                                                 sneakpeek)
                                                 
    if not os.isatty(0):
        print('Execution complete. Time elapsed: '+str(datetime.datetime.now() - begin_time))
    else:
        input('Execution complete. Time elapsed: '+str(datetime.datetime.now() - begin_time)+'\nPress Enter to exit.')
#here multithreaded prints execution of main()